#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" MCNP HTML Output Writer
Usage: python3 mcnp_reader.py [-o mcnp_filename.out] [-n scaling_factor]

This programme takes an MCNP output file and creates a HTML summary of the results.
If run from the command line, it can take two optional arguments; the
mcnp filename and a normalisation factor. if the mcnp filename is not supplied, a tkinter
window will open to allow a file to be selected. If a normalisation factor is not applied,
it will be assumed to be 1.

"""

# Imports from standard library
import re
# Local imports
from . import particles, mcnp_html_writer, tallies, global_variables as gv, cells


def read_file(filename):
    """Read in the MCNP output file.

    Args:
        filename (str): The name (and location) of the MCNP .out file

    Returns:
        data (list): The lines from the output file
    """
    with open(filename, 'r') as file:
        data = file.readlines()
    return data


def parse_output(output_data):
    """Send the MCNP data to functions in various modules, which extract the relevant data for themselves

    Args:
        output_data (list): The MCNP output data

    Returns:
        None
    """
    gv.date_time = get_date_time(output_data)
    get_runtime(output_data)
    gv.mcnp_input = get_input(output_data)
    gv.parameters = get_parameters(gv.mcnp_input)
    cell_data = cells.find_cells(output_data)
    neutron_populations = cells.get_particle_populations(output_data, 'neutron')
    photon_populations = cells.get_particle_populations(output_data, 'photon')
    electron_populations = cells.get_particle_populations(output_data, 'electron')
    cells.create_cell_objects(cell_data)
    for cell in gv.cell_list:
        cell.assign_populations(neutron_populations, photon_populations, electron_populations)
    get_warnings(output_data)
    get_comments(output_data)
    get_duplicate_surfaces(output_data)
    particles.get_neutrons(output_data)
    particles.get_photons(output_data)
    particles.get_electrons(output_data)
    if gv.crit_case:
        gv.k_effective = get_k_eff(output_data)
        get_active_cycles(output_data)
    else:
        tallies.get_tallies(output_data)
        for tally in gv.tally_list:
            tally.normalise_data()
    # tallies.Tally.sort_for_html()


def get_date_time(output_data):
    """Get the date and time that the mcnp case was run

    Args:
        output_data (list): The MCNP output file

    Returns:
        date_time (dict): the date and time the case was run
    """
    for line in output_data:
        if line.startswith('1mcnp'):
            time = line.split()[5]
            date = line.split()[4]
            date = date.split('/')
            d = date[1]
            m = date[0]
            y = '20' + date[2]
            date_time = {'date': f"{y}/{m}/{d}", 'time': time}
            gv.rundate = f"{y}/{m}/{d}"
            gv.runtime = time
            return date_time


def get_runtime(output_data):
    """Get the time or number of particles the MCNP case was run for

    Args:
        output_data (list): The MCNP output file

    Returns:
        None, but writes to either or both of gv.ctme or gv.nps project-level variables
    """
    pattern_nps = re.compile(r'^\s{6,}\d+-\s{7}(nps|NPS)\s+.+')
    pattern_ctme = re.compile(r'^\s{6,}\d+-\s{7}(ctme|CTME)\s+.+')
    for line in output_data:
        if pattern_ctme.match(line):
            gv.ctme = line.split()[2]
        if pattern_nps.match(line):
            gv.nps = line.split()[2]


def get_input(output_data):
    """Extract the MCNP input deck from the MCNP output file

    Args:
        output_data (list): The MCNP output data

    Returns:
        mcnp_input (list): The MCNP input deck
    """
    pattern_input = re.compile(r'^\s{6,}\d+-\s{7}(.+)')
    mcnp_input = []
    for line in output_data:
        match = pattern_input.match(line)
        if match:
            mcnp_input.append(match.group(1).strip('\n'))
    return mcnp_input


def get_parameters(input_data):
    """Get any parameters used in the input file.
    NOTE: This is only useful if MCNP is run using the Cerberus
    package CYCLONE; non-cyclone users can ignore this function.

    Args:
        input_data (list): The MCNP input file

    Returns:
        variables (dict): All the parameters used by this case
    """
    variables = {}
    for n, line in enumerate(input_data):
        if "USING THE FOLLOWING VARIABLES" in line:
            for m, other_line in enumerate(input_data[n+1:], start=n+1):
                if '=' not in other_line:
                    break
                variable = other_line.split()[1]
                value = other_line.split()[3]
                variables[variable] = float(value)
            break
    return variables


def get_warnings(output_data):
    """Find the warnings in the MCNP output data

    Args:
        output_data (list): The MCNP output data

    Returns:
        None, but populates gv.warnings, a list of the warning messages
    """
    PATTERN_warnings = re.compile(r'warning')
    for line in output_data:
        if PATTERN_warnings.search(line):
            if "warning message so far" not in line:
                warning = line[10:].strip().capitalize()
                if warning not in gv.warnings:
                    gv.warnings.append(warning)


def get_comments(output_data):
    """Find the comments in the MCNP output file.

    Args:
        output_data (list): The mcnp output

    Returns:
        None, but populates gv.comments, a list of the comments
    """
    PATTERN_comments = re.compile(r'comment\.\s+[A-Za-z0-9].+')    # Ignores blank comment lines
    for line in output_data:
        if PATTERN_comments.search(line):
            comment = line[10:].strip().capitalize()
            gv.comments.append(comment)


def get_duplicate_surfaces(output_data):
    """Find all the duplicate surface messages in the mcnp output

    Args:
        output_data (list): The mcnp output

    Returns:
        None, but populates gv.duplicate_surfaces, a list of the duplicate surfaces
    """
    PATTERN_duplicates = re.compile(r'\ssurface\s+\d+.+and surface.+are the same.+')
    for line in output_data:
        if PATTERN_duplicates.match(line):
            if line not in gv.duplicate_surfaces:
                gv.duplicate_surfaces.append(line)


def get_k_eff(output_data):
    """Search the output data for the section concerning k-effective, and create a single dictionary holding the data

    Args:
        output_data (list): The MCNP output

    Returns:
         k_eff (dict): the k-effective values for the first half, second half and total run
    """
    PATTERN_k_eff = re.compile(r'^\s*problem\s+keff.+')
    k_eff = {}
    for num, line in enumerate(output_data):
        if PATTERN_k_eff.match(line):
            first_half = re.split(r'\s{2,}', output_data[num+2].strip())
            k_eff['first half k_eff'] = first_half[1]
            k_eff['first half stdev'] = first_half[2]

            second_half = re.split(r'\s{2,}', output_data[num+3].strip())
            k_eff['second half k_eff'] = second_half[1]
            k_eff['second half stdev'] = second_half[2]

            final_result = re.split(r'\s{2,}', output_data[num+4].strip())
            k_eff['final k_eff'] = final_result[1]
            k_eff['final stdev'] = final_result[2]
    return k_eff


def get_active_cycles(output_data):
    """For a crit case, find the number of active and inactive cycles

    Args:
        output_data (list): the output file

    Returns:
        None, but writes cycles (dict) to gv.cycles

    """
    cycles = {}
    for line in output_data:
        if "the minimum estimated standard deviation for the col/abs/tl keff estimator occurs with" in line:
            cycles["inactive"] = int(line.split()[12])
            cycles["active"] = int(line.split()[16])
    gv.cycles = cycles


def main(filename, scaling_factor=1, crit_case=False):
    """ Provide entry point to this module and call the other parts of the script

    Args:
        filename (str): the mcnp output file
        scaling_factor (float): A number which all results will be multiplied by
        crit_case (bool): True if criticality case, False (default) for shielding case

    Returns:
        None
    """
    gv.scaling_factor = scaling_factor
    gv.crit_case = crit_case
    output_data = read_file(filename)
    parse_output(output_data)
    mcnp_html_writer.main(filename)



#!/usr/bin/python3
# Cerberus Nuclear Ltd

"""
    This module creates all the CSS used in the HTML output
"""


def i(arg):
    """ Indent the arg string to make the html more readable """
    return f'   {arg}'


def reset():
    """ CSS reset to remove any inbuilt CSS in the web browser. """
    return """
        /*
        html5doctor.com Reset Stylesheet
        v1.6.1
        Last Updated: 2010-09-17
        Author: Richard Clark - http://richclarkdesign.com
        Twitter: @rich_clark
        */

        html, body, div, span, object, iframe,
        h1, h2, h3, h4, h5, h6, p, blockquote, pre,
        abbr, address, cite, code,
        del, dfn, em, img, ins, kbd, q, samp,
        small, strong, sub, sup, var,
        b, i,
        dl, dt, dd, ol, ul, li,
        fieldset, form, label, legend,
        table, caption, tbody, tfoot, thead, tr, th, td,
        article, aside, canvas, details, figcaption, figure,
        footer, header, hgroup, menu, nav, section, summary,
        time, mark, audio, video {
            margin:0;
            padding:0;
            border:0;
            outline:0;
            font-size:100%;
            vertical-align:baseline;
            background:transparent;
        }

        body {
            line-height:1;
        }

        article,aside,details,figcaption,figure,
        footer,header,hgroup,menu,nav,section {
            display:block;
        }

        nav ul {
            list-style:none;
        }

        blockquote, q {
            quotes:none;
        }

        blockquote:before, blockquote:after,
        q:before, q:after {
            content:'';
            content:none;
        }

        a {
            margin:0;
            padding:0;
            font-size:100%;
            vertical-align:baseline;
            background:transparent;
        }

        /* change colours to suit your needs */
        ins {
            background-color:#ff9;
            color:#000;
            text-decoration:none;
        }

        /* change colours to suit your needs */
        mark {
            background-color:#ff9;
            color:#000;
            font-style:italic;
            font-weight:bold;
        }

        del {
            text-decoration: line-through;
        }

        abbr[title], dfn[title] {
            border-bottom:1px dotted;
            cursor:help;
        }

        table {
            border-collapse:collapse;
            border-spacing:0;
        }

        /* change border colour to suit your needs */
        hr {
            display:block;
            height:1px;
            border:0;
            border-top:1px solid #cccccc;
            margin:1em 0;
            padding:0;
        }

        input, select {
            vertical-align:middle;
        }

        /* End of CSS Reset */
        /*------------------------------------------------*/

        """


def write_css(f):
    """ Write standard CSS for the HTML output. """
    # HTML CSS STYLING FOR SECTIONS
    f.write(i('<style type="text/css">\n'))
    f.write(reset())
    f.write(i('\n\n'))

    # general
    f.write(i('html, body {height: 100%; background-color: white; font-family: "Work Sans", Raleway, sans-serif;}\n'))
    f.write(i('p {text-align: center; color: #333333}\n'))
    f.write(i('h1 {text-align: center; padding: 10px 0px; color: #1c76bc; font-size: 28px; margin-top: 20px;}\n'))
    f.write(i('h2 {text-align: center; padding: 10px 0px; color: #333333; font-size: 20px;}\n'))
    f.write(i('h3 {text-align: center; padding: 10px 0px; color: #333333;; font-size: 16px;}\n'))
    f.write(i('div {padding: 20px 0px;}\n'))
    f.write(i('div .tally-results {padding: 5px 0px;}\n'))
    f.write(i('div .tally-checks {padding: 5px 0px;}\n'))
    f.write(i('.case-details {font-size: 16px}\n'))

    # tables
    f.write(i('.table {text-align:center; border-collapse:collapse; margin:auto; line-height:10px; border:1px solid black;}\n'))
    f.write(i('.table td {padding:10px; line-height: 120%; font-size:14px; color: #333333; border:1px solid black;}\n'))
    f.write(i('.table th {padding:10px; line-height: 120%; font-size:14px; color: #333333; border:1px solid black; background-color: #cce0ff;  }\n'))

    # navbar
    f.write(i('''.navbar {
        position: fixed;
        background-color: #cce0ff;
        margin: 0;
        border-bottom: 1px solid #1c76bc ;
        width: 100%; 
        height: auto;
        padding: 5px;
        }\n'''))
    f.write(i('''.navbar ul {
        list-style: none;
        display: flex; 
        flex-direction: row; 
        flex-wrap: wrap; 
        justify-content: center; 
        }'''))
    f.write(i('''.navbar li a {
        padding: 5px 20px;
        line-height: 170%;
        text-align: center;
        text-decoration: none;
        font-size: 16px;
        color: rgba(51,51,51,1);
        }\n'''))
    f.write(i('.navbar a:hover {background-color: #1c76bc; color: #FFFFFF}\n'))

    # cerberus logo
    f.write(i('#logo {text-align: center; padding-top: 100px;} \n'))
    f.write(i('#logo img{width: 50%; height: auto;} \n'))


    # details
    f.write(i('#details .table {border: none; padding: 10px; table-layout: fixed; width: 100%;}'))
    f.write(i('#details td {border: none; padding: 10px; font-size: 20px; width: 50%;} \n'))

    # parameters
    f.write(i('#parameters table {border: none; padding: 10px; table-layout: fixed; width: 100%;} \n'))
    f.write(i('#parameters td {border: none; padding: 5px 10px; font-size: 20px; line-height: 120%; width: 50%;} \n'))


    # warnings
    f.write(i('#warnings, #duplicates, #comments {text-align: center;}\n'))
    f.write(i('#warning_messages, #duplicate_messages, #comment_messages {display:inline-block; padding: 5px}\n'))
    f.write(i('#warning_messages, p {text-align: left; padding: 5px; }\n'))
    f.write(i('#duplicate_messages, p {text-align: left; padding: 5px; }\n'))

    # comments
    #f.write(i('#comments {text-align: center;}\n'))
    #f.write(i('#comment_messages {display:inline-block;}\n'))
    f.write(i('#comment_messages p {text-align: left; padding: 5px;}\n'))

    # mcnp input
    f.write(i('#input_file {text-align: center;}\n'))
    f.write(i('#input {display:inline-block; }\n'))
    f.write(i('#input p {white-space:pre; font-family: Courier,monospace; font-weight:bold; text-align:left; }\n'))
    f.write(i('</style>\n'))



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""Eddy is a programme to convert the text-based output files from MCNP and SCALE to
a user-friendly HTML format.

Copyright (C) 2020  Cerberus Nuclear Ltd

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


This module is the entry point for Eddy, the MCNP and SCALE to HTML output converter.
If run as __main__, this script asks for an output file and scaling factor, 
determines whether it is an MCNP or SCALE case, and calls either mcnp.mcnp_converter.py 
or scale.scale_converter.py. 
Alternatively, main() can be called directly by another module and provided with the same two arguments.
"""

# Imports from standard library
import sys
import argparse
from tkinter import Tk, simpledialog
from tkinter.filedialog import askopenfilename
# Local imports
# none

if __name__ == "__main__":
    from scale import scale_converter
    from mcnp import mcnp_converter
else:
    from .scale import scale_converter
    from .mcnp import mcnp_converter


def check_if_crit(output_data):
    """Read through output data to check if case is crit or shielding

    Args:
        output_data (list): The text of the output file

    Returns:
        True if kcode found or False if not.

    """
    for line in output_data:
        if 'kcode' in line.lower():
            return True
    return False


def read_file(filename):
    """Read in the output file.

    Args:
        filename (str): The name (and location) of the .out file

    Returns:
        data (list): The lines from the output file
    """
    with open(filename, 'r') as file:
        data = file.readlines()
    return data


def get_filename(args=None):
    """ Get the name and path of the MCNP output file
    
    Args: 
        (optional) args: the argparse arguments, if they exist
    
    Returns:
        filename (str): The filename including filepath
    """
    if args.file:               # check if filename was passed as a command line argument
        filename = args.file
    else:                       # open tkinter window for file selection
        Tk().withdraw()
        filename = askopenfilename(
            title="Select Output File",
            filetypes=(("output files", "*.out"), ("all files", "*.*"))
            )
    return filename


def get_scaling_factor(args=None):
    """ Get the scaling factor to multiply results by for a shielding case
    Args: 
        (optional) args: the argparse arguments, if they exist
    Returns:
        scaling_factor (float): The scaling factor
    """
    if args.scaling_factor:                         # check if scaling factor was passed as a command line argument 
        scaling_factor = args.scaling_factor
    else:                                           # open tkinter window to ask for scaling factor
        Tk().withdraw()
        scaling_factor = simpledialog.askfloat(
        title="Scaling Factor",
        prompt="Please enter a scaling factor to multiply your results by", initialvalue=1,
        )
    # type checking for scaling factor
    if type(scaling_factor) is not float:
        try:
            scaling_factor = float(scaling_factor)
        except ValueError:
            print("The scaling factor should be a floating-point (decimal) number")
            sys.exit()
    return scaling_factor


def get_args(filename=None, scaling_factor=None):
    """
    Get the args from argparse if they are sent by the command line,
    Call functions to get filename and scaling factor.
    Determine if this is a crit case (for MCNP).
    Read in the output data from the mcnp output file.

    Args:
        filename (optional) (str or None): the file path (including the name) of the output file, may be None
        scaling_factor (optioonal) (float or None): the scaling factor to multiply results by, may be None

    Returns:
        filename (str): the file path (including the name) of the output file
        output_data (list): The contents of the output file
        scaling_factor (float): the scaling factor to multiply results by
        crit_case (bool): True if kcode case, otherwise False
    """
    parser = argparse.ArgumentParser(description='MCNP or SCALE output to HTML Converter')
    parser.add_argument("-o", "--file", help="MCNP or SCALE output file")
    parser.add_argument("-sf", "--scaling_factor", type=float, help="Scaling Factor")
    args = parser.parse_args()

    # Get filename (and path) of mcnp output
    if filename is None:            # if filename was not passed as an argument to main()...
        filename = get_filename(args)
    # check that a file was actually selected:
    assert len(filename) != 0, "No file was selected."
    output_data = read_file(filename)

    # Check if shielding or crit case
    crit_case = check_if_crit(output_data)

    # Ask for scaling factor for shielding cases
    if crit_case:
        # set scaling factor to 1 for crit cases (this variable will not be used)
        scaling_factor = 1
    else:
        if scaling_factor is None:
            scaling_factor = get_scaling_factor(args)
       
    print(f"Output file: {filename}")
    return filename, output_data, scaling_factor, crit_case


def main(filename=None, scaling_factor=None):
    """Entry point to Eddy. Can take filename and scaling factor as arguments.
    Call get_args to find the filename & scaling factor if not provided, and also get the
    output data and determine whether it is a crit case.
    Call read_file() and determines whether it is an MCNP or SCALE case;
    call the relevant converter.

    Args:
        filename (optional) (str): the file path (including the name) of the output file
        scaling_factor (optional) (float): A number by which the results will be multiplied
    """
    filename, output_data, scaling_factor, crit_case = get_args(filename, scaling_factor)
    if 'SCALE' in output_data[2]:
        scale_converter.main(filename, scaling_factor)
    elif 'MCNP' in output_data[0]:
        mcnp_converter.main(filename, scaling_factor, crit_case)
    else:
        raise RuntimeError("This file doesn't seem to be an MCNP or SCALE output?")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" This module creates the HTML output for the SCALE data """

# Standard library imports
import os
import datetime
import pkgutil
# Third party imports
from jinja2 import Template
#local imports
from . import scale_global_variables as gv


def get_css():
    """Get the css to be used in the html output

    Returns:
        inline_css (str): The whole static css file as a single string
    """
    inline_css = pkgutil.get_data(__name__, "../static/style.css").decode("utf-8")
    return inline_css


def sanitize_input(scale_input):
    """Replaces any < and > characters in the scale output with html codes &lt and &gt to stop
    them from being interpreted as html tags.

    Args: mcnp_input [list]: The mcnp input file
    Returns: san_mcnp_input [list]: The sanitized version of the mcnp input
    """
    san_scale_input = []
    for line in scale_input:
        if "<" in line:
            line = line.replace("<","&lt")
        if ">" in line:
            line = line.replace(">", "&gt")
        san_scale_input.append(line)
    return san_scale_input


def get_html(filename, inline_css):
    """Get the html output from the jinja template

    Args:
        filename (str): The file path of the SCALE output file
        inline_css (str): The css as a single string

    Returns:
        html (str): the html output as a single string
    """
    scale_input = sanitize_input(gv.scale_input)
    html_template = pkgutil.get_data(__name__, "../static/SCALE_template.html").decode("utf-8")
    template = Template(html_template)
    case_name, extension = os.path.splitext(filename)
    case_name = case_name.replace('\\', '/').split('/')[-1]
    # render template as a unicode string
    html = template.render(
        filename=filename,
        case_name=case_name,
        #logo=f"{os.getcwd()}/static/logo.png",
        inline_css=inline_css,
        rundate=gv.rundate,
        runtime=gv.runtime,
        date=datetime.datetime.now().strftime("%Y/%m/%d"),
        time=datetime.datetime.now().strftime("%H:%M:%S"),
        scaling_factor=gv.scaling_factor,
        tally_list=gv.tally_list,
        mixture_list=gv.mixture_list,
        input_deck=scale_input,
        )
    return html


def main(filename):
    """Provide entry point to this module and call other functions to create HTML,
    write that completed html to the output file

    Args:
        filename (str): The file path of the SCALE output file

    Returns:
        None
    """
    inline_css = get_css()
    output = get_html(filename, inline_css)
    output_file, extension = os.path.splitext(filename)
    output_file += '.html'
    with open(output_file, 'w') as f:
        f.write(output)





#!/usr/bin/python3
# Peter Evans
# Cerberus Nuclear Ltd

"""
This module holds the code that writes HTML to the html file.
"""


# Imports from standard library
import datetime
import os
import pkgutil
# Third party imports:
from jinja2 import Template
# local imports
from . import global_variables as gv


def get_css():
    """Read in css from static css file

    Returns:
        inline_css (str): the whole static css file
    """
    inline_css = pkgutil.get_data(__name__, "../static/style.css").decode("utf-8")
    return inline_css


def sanitize_input(mcnp_input):
    """Replaces any < and > characters in the mcnp output with html codes &lt and &gt to stop
    them from being interpreted as html tags.

    Args: mcnp_input [list]: The mcnp input file
    Returns: san_mcnp_input [list]: The sanitized version of the mcnp input
    """
    san_mcnp_input = []
    for line in mcnp_input:
        if "<" in line:
            line = line.replace("<","&lt")
        if ">" in line:
            line = line.replace(">", "&gt")
        san_mcnp_input.append(line)
    return san_mcnp_input


def get_html(filename, inline_css):
    """Get the HTML from the jinja template

    Args:
        filename (str): The mcnp output file path
        inline_css (str): The whole css file as a single string

    Returns:
        html (str): The completed html output
    """
    case_name, extension = os.path.splitext(filename)
    case_name = case_name.replace('\\', '/').split('/')[-1]
    mcnp_input = sanitize_input(gv.mcnp_input)
    #with open("static/MCNP_template.html", "r") as file:
    #    html_template = file.read()
    html_template = pkgutil.get_data(__name__, "../static/MCNP_template.html").decode("utf-8")

    template = Template(html_template)
    html = template.render(
        filename=filename, # with path
        case_name=case_name,
        #logo=f"{os.getcwd()}/static/logo.png",
        inline_css=inline_css,
        rundate=gv.rundate,
        runtime=gv.runtime,
        date=datetime.datetime.now().strftime("%Y/%m/%d"),
        time=datetime.datetime.now().strftime("%H:%M:%S"),
        scaling_factor=gv.scaling_factor,
        parameters=gv.parameters,
        k_eff=gv.k_effective,
        tally_list=gv.tally_list,
        tally_types=gv.f_types,
        f2_tallies=gv.F2_tallies,
        f4_tallies=gv.F4_tallies,
        f5_tallies=gv.F5_tallies,
        warnings=gv.warnings,
        comments=gv.comments,
        duplicate_surfaces=gv.duplicate_surfaces,
        cell_list=gv.cell_list,
        particle_list=gv.particle_list,
        input_deck=mcnp_input,
        cycles=gv.cycles,
        )
    return html


def main(filename):
    """Provide entry point to this module and call other functions to create HTML,
    write that completed html to the output file

    Args:
        filename (str): The file path of the MCNP output file

    Returns:
        None
    """
    css = get_css()
    output = get_html(filename, css)
    output_file, extension = os.path.splitext(filename)
    output_file += '.html'
    with open(output_file, 'w') as f:
        f.write(output)



#!/usr/bin/python3
# Peter Evans
# Cerberus Nuclear Ltd
# November 2019

"""
    This module holds all the code related to cells, including the Cells class,
    and the functions which find the cell data in the mcnp output, extract the cell
    section of that data, and send it to the Cell class to create the Cell objects.

    It also contains the code to find the neutron, photon and electron population data,
    which is assigned to each cell object as they are created.
"""
# imports from standard library
import re
from pathlib import Path
# local imports
from . import global_variables as gv


class Cell:
    """ This class holds the attributes and methods for Cell objects."""
    # pylint: disable=too-many-instance-attributes
    # MCNP cells have a lot of attributes

    def __init__(self, **info):
        self.cell_number = info['cell_number']
        self.material_number = info['material_number']
        self.atom_density = info['atom_density']
        self.gram_density = info['gram_density']
        self.volume = float(info['volume'])
        self.mass = float(info['mass'])
        self.neutron_importance = info['neutron_importance']
        self.photon_importance = info['photon_importance']
        self.electron_importance = info['electron_importance']
        gv.cell_list.append(self)

    def assign_populations(self, neutron_pop=None, photon_pop=None, electron_pop=None):
        """Assign neutron, photon and electron population data to each cell.

        Args:
            neutron_pop (list): The neutron population section of the MCNP output
            photon_pop (list): The photon population section of the MCNP output
            electron_pop (list): The electron population section of the MCNP output

        Returns:
            None. The data is assigned to the cell objects as various new attributes.
        """
        # Assign neutron population data to cell
        if neutron_pop:
            for line in neutron_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.neutron_tracks_entering = line[2]
                    self.neutron_population = line[3]
                    self.neutron_collisions = line[4]
        # Assign photon population data to cell
        if photon_pop:
            for line in photon_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.photon_tracks_entering = line[2]
                    self.photon_population = line[3]
                    self.photon_collisions = line[4]
        # Assign electron population data to cell
        if electron_pop:
            for line in electron_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.electron_tracks_entering = line[2]
                    self.electron_population = line[3]
                    self.electron_collisions = line[4]

    def describe_object(self):
        """Print a description of the cell object to the terminal (for debugging purposes)

        Args:
            self, the cell object

        Returns:
            None
        """
        print(f"Cell number: {self.cell_number}")
        print(f"Cell material: {self.material_number}")
        print(f"Cell density: {self.gram_density} g/cm3")
        print(f"Cell volume: {self.volume} cm3")
        # The following attributes are particle-dependent, and so may or may not exist for any one mcnp case
        # pylint: disable=multiple-statements
        try:
            print(f"Neutron Importance: {self.neutron_importance}")
        except AttributeError:
            pass
        try:
            print(f"Photon Importance: {self.photon_importance}")
        except AttributeError:
            pass
        try:
            print(f"Electron Importance: {self.electron_importance}")
        except AttributeError:
            pass
        print()


############################################################
#  End of Cell class                                       #
############################################################


def find_cells(output_data):
    """Loop through the mcnp output to find the section containing the cell data

    Args:
        output_data (list): the mcnp output

    Returns:
        cell_section (list): The section of the output containing the cell data
    """
    PATTERN_cells_start = re.compile(r'1cells')
    PATTERN_cells_ends = re.compile(r' total')
    cell_section = []
    for line in output_data:
        if PATTERN_cells_start.match(line):
            cell_start_line = output_data.index(line)
            for row in output_data[cell_start_line:]:
                cell_section.append(row.strip('\n'))
                if PATTERN_cells_ends.match(row):
                    break
            break
    return cell_section


def create_cell_objects(data):
    """Take the cell data and create objects of the Cell class

    Args:
        data (list): a list of strings containing the cell part of the mcnp output

    Returns:
        None: creates Cell objects
    """
    header = (data[2].split())
    PATTERN_cell = re.compile(r'\s+\d+\s.+')
    for line in data[1:]:
        if PATTERN_cell.match(line):
            info = line.split()
            neutron_importance = get_particle_importance(header, info, 'neutron')
            photon_importance = get_particle_importance(header, info, 'photon')
            electron_importance = get_particle_importance(header, info, 'electron')
            Cell(
                cell_number=info[1],
                material_number=info[2],
                atom_density=info[3],
                gram_density=info[4],
                volume=info[5],
                mass=info[6],
                neutron_importance=neutron_importance,
                photon_importance=photon_importance,
                electron_importance=electron_importance
                )


def get_particle_importance(header, data, particle):
    """For a particular particle type, determine the importance of that particle.
    This function is called for each cell as part of its creation.

    Args:
        header (list): the particle population table header
        data (list): the section of the mcnp output with the cell data
        particle (str): either 'neutron', 'photon' or 'electron'

    Returns:
         particle_importance (str): the particle importance, or 0.
    """
    if particle in header:
        particle_importance = data[header.index(particle) + 6]
    else:
        particle_importance = '0'
    return particle_importance


def get_particle_populations(output_data, particle):
    """Find and return the section of the output data concerning populations of a particular particle

    Args:
        output_data (list): the mcnp output
        particle (str): either 'photon', 'neutron' or 'electron'

    Returns:
        particle_populations (list): the lines from the output file concerning populations of that particle
    """
    PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
    PATTERN_particle_populations = re.compile(fr'^1{particle}\s+activity\sin\seach\scell.+')
    PATTERN_end_populations = re.compile(r'\s+total.+')
    particle_populations = None
    for n, line in enumerate(output_data):
        if PATTERN_run_terminated.match(line):
            terminated = n
            for m, new_line in enumerate(output_data[terminated:], start=terminated):
                if PATTERN_particle_populations.match(new_line):
                    for p in range(m, len(output_data)):
                        if PATTERN_end_populations.match(output_data[p]):
                            particle_populations = output_data[m:p + 1]
                            return particle_populations
    return None



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" This module holds the code related to particle data. """

# Standard library imports
import re
# local imports
from . import global_variables as gv


class Particle:
    """
    This class exists to hold the data on particle populations, creation and loss.
    """

    def __init__(self, creation_data, loss_data, total_data, particle_type):
        gv.particle_list.append(self)
        self.attributes = []
        self.particle_type = particle_type
        for item in creation_data:
            name = ' '.join(item.split()) + ' creation'
            name = name.title()
            self.__dict__[name] = creation_data[item]
            self.attributes.append(name)
        for item in loss_data:
            name = ' '.join(item.split()) + ' loss'
            name = name.title()
            self.__dict__[name] = loss_data[item]
            self.attributes.append(name)
        for item in total_data:
            name = ' '.join(item.split()).title()
            self.__dict__[name] = total_data[item]
            self.attributes.append(name)

    def describe_object(self):
        """ Print the attributes of the particle to the terminal."""
        print()
        print(f"Printing {self.particle_type} Data...".upper())
        for item in self.attributes:
            print(f"{item}: {self.__dict__[item]}")
        print()

# ----------------------------------------------------------------------------------
# NEUTRONS -------------------------------------------------------------------------
# ----------------------------------------------------------------------------------


def get_neutrons(output_data):
    """Find the mcnp neutron data then create a Particle object with that data

    Args:
        output_data (list): The MCNP output

    Returns:
        None
    """
    neutron_data = find_mcnp_neutron_data(output_data)
    if neutron_data:
        creation_data, loss_data, total_data = sort_mcnp_neutron_data(neutron_data)
        Particle(creation_data, loss_data, total_data, 'neutron')


def find_mcnp_neutron_data(output_data):
    """Find the neutron data in the mcnp output.

    Args:
        output_data (list): The MCNP output

    Returns:
        neutron_data (list): The part of the MCNP output with the neutron particle data
    """
    PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
    PATTERN_neutron = re.compile(r'\sneutron\screation\s+tracks.+')
    neutron_data = None
    # Find the neutron data and save to a list
    for n, line in enumerate(output_data):
        if PATTERN_run_terminated.match(line):
            terminated = n
            for m, new_line in enumerate(output_data[terminated:], start=terminated):
                if PATTERN_neutron.match(new_line):
                    neutron_data = output_data[m:m+28]
                    break
    return neutron_data


def sort_mcnp_neutron_data(data):
    """Separate out the neutron data into creation, loss, and total data,
    then split each by indices and put into a dictionary

    Args:
        data (list): The part of the MCNP output with the neutron particle data

    Returns:
        a tuple containing
            creation_dict: The neutron creation data,
            loss_dict: The neutron loss data,
            total_dict: The total neutron data,
    """
    indices = [0, 18, 27, 41, 55]   # These have been put in based on inspection of an example output file
    creation_dict = {}
    loss_dict = {}
    for line_num in range(3, 22):
        # Separate neutron creation data
        creation = data[line_num][:55]
        creation = [creation[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        creation_dict[creation[0]] = {'tracks': int(creation[1]),
                                      'weight': float(creation[2]),
                                      'energy': float(creation[3]),
                                      }
        # Separate neutron loss data
        loss = data[line_num][64:]
        loss = [loss[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        loss_dict[loss[0]] = {'tracks': int(loss[1]),
                              'weight': float(loss[2]),
                              'energy': float(loss[3]),
                              }
    total_dict = {}
    for line_num in range(23, 27):
        total = data[line_num][:42].strip(), data[line_num][42:53].strip()
        total_dict[total[0]] = total[1]
    return creation_dict, loss_dict, total_dict


# ----------------------------------------------------------------------------------
# PHOTONS --------------------------------------------------------------------------
# ----------------------------------------------------------------------------------


def get_photons(output_data):
    """Find the mcnp photon data then create a Particle object with that data

    Args:
        output_data (list): The MCNP output

    Returns:
        None
    """
    photon_data = find_mcnp_photon_data(output_data)
    if photon_data:
        creation_data, loss_data, total_data = sort_mcnp_photon_data(photon_data)
        Particle(creation_data, loss_data, total_data, 'photon')


def find_mcnp_photon_data(output_data):
    """Find the photon data in the mcnp output

    Args:
        output_data (list): The MCNP output

    Returns:
        photon_data (list): The part of the MCNP output with the photon particle data
    """
    PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
    PATTERN_photon = re.compile(r'\sphoton\screation\s+tracks.+')
    photon_data = None
    for n, line in enumerate(output_data):
        if PATTERN_run_terminated.match(line):
            terminated = n
            for m, new_line in enumerate(output_data[terminated:], start=terminated):
                if PATTERN_photon.match(new_line):
                    photon_data = output_data[m:m+33]
                    break
    return photon_data


def sort_mcnp_photon_data(photon_data):
    """Separate out the photon data into creation, loss, and total data,
    then split each by indices and put into a dictionary

    Args:
        photon_data (list): The part of the MCNP output with the photon particle data

    Returns:
        a tuple containing
            creation_dict: The photon creation data,
            loss_dict: The photon loss data,
            total_dict: The photon neutron data,
    """
    indices = [0, 18, 27, 41, 55]
    creation_dict = {}
    loss_dict = {}
    for line_num in range(3, 27):
        # Separate photon creation data
        creation = photon_data[line_num][:55]
        creation = [creation[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        creation_dict[creation[0]] = {'tracks': int(creation[1]),
                                      'weight': float(creation[2]),
                                      'energy': float(creation[3]),
                                      }
        # Separate photon loss data
        loss = photon_data[line_num][64:]
        loss = [loss[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        try:
            loss_dict[loss[0]] = {'tracks': int(loss[1]), 'weight': float(loss[2]), 'energy': float(loss[3])}
        except ValueError:
            # There are more creation lines than loss lines
            pass

    total_dict = {}
    for line_num in range(28, 32):
        total = photon_data[line_num][:42].strip(), photon_data[line_num][42:53].strip()
        total_dict[total[0]] = total[1]
    # for item in total_dict: print(item, total_dict[item])
    return creation_dict, loss_dict, total_dict

# ----------------------------------------------------------------------------------
# ELECTRONS ------------------------------------------------------------------------
# ----------------------------------------------------------------------------------


def get_electrons(output_data):
    """Find the mcnp electron data then create a Particle object with that data

    Args:
        output_data (list): The MCNP output

    Returns:
        None
    """
    electron_data = find_mcnp_electron_data(output_data)
    if electron_data:
        creation_data, loss_data, total_data = sort_mcnp_electron_data(electron_data)
        Particle(creation_data, loss_data, total_data, 'electron')


def find_mcnp_electron_data(output_data):
    """Find the electron data in the mcnp output

    Args:
        output_data (list): The MCNP output

    Returns:
        electron_data (list): The part of the MCNP output with the electron particle data
    """
    PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
    PATTERN_electron = re.compile(r'\selectron\screation\s+tracks.+')
    electron_data = None
    for n, line in enumerate(output_data):
        if PATTERN_run_terminated.match(line):
            terminated = n
            for m, new_line in enumerate(output_data[terminated:], start=terminated):
                if PATTERN_electron.match(new_line):
                    electron_data = output_data[m:m+24]
                    break
    return electron_data


def sort_mcnp_electron_data(electron_data):
    """Separate out the electron data into creation, loss, and total data,
    then split each by indices and put into a dictionary

    Args:
        electron_data (list): The part of the MCNP output with the electron particle data

    Returns:
        a tuple containing
            creation_dict: The electron creation data,
            loss_dict: The electron loss data,
            total_dict: The electron neutron data,
    """
    indices = [0, 18, 27, 41, 55]
    creation_dict = {}
    loss_dict = {}
    for line_num in range(3, 18):
        # Separate electron creation data
        creation = electron_data[line_num][:55]
        creation = [creation[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        creation_dict[creation[0]] = {'tracks': int(creation[1]),
                                      'weight': float(creation[2]),
                                      'energy': float(creation[3]),
                                      }
        # Separate electron loss data
        loss = electron_data[line_num][64:]
        loss = [loss[indices[i]:indices[i+1]].strip() for i in range(len(indices)-1)]
        try:
            loss_dict[loss[0]] = {'tracks': int(loss[1]), 'weight': float(loss[2]), 'energy': float(loss[3])}
        except ValueError:
            pass  # ignore single empty entry

    total_dict = {}
    for line_num in range(19, 23):
        total = electron_data[line_num][:42].strip(), electron_data[line_num][42:53].strip()
        total_dict[total[0]] = total[1]
    return creation_dict, loss_dict, total_dict



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""
    This module holds all the code relating to tallies.
"""

# standard library imports
import re
# local imports:
from . import global_variables as gv


class Tally:
    """Each mcnp tally is represented by a Tally object containing all the available data about that tally.
    All tallies should be members of one of the subclass of this Tally class.
    """

    def __init__(self, data):
        # for item in data: print(item)
        self.data = data    # can be removed once this method is complete
        self.tally_number = data[0].split()[1]
        self.nps = data[0].split()[4]
        self.f_type = 'F' + data[1].split()[2]
        self.tally_type = data[1].split()[3:]
        self.tally_type = ' '.join(self.tally_type)
        self.particles = data[2].split()[1]
        self.dose_functions = data[3].split()[7], data[3].split()[9]
        self.results = self.get_results(data)
        self.statistical_checks = self.get_statistical_checks(data)
        self.passes = self.get_passes()
        gv.tally_list.append(self)
        if self.f_type not in gv.f_types:
            gv.f_types.append(self.f_type)

    def get_results(self, data):
        """This is a placeholder method; each subclass of Tally should have its own 'get_results' method"""
        raise NotImplementedError(f"The {self.__class__} subclass should have its own normalise_data method")

    def get_statistical_checks(self, data):
        """Get the statistical check results from the mcnp output file

        Args:
            data (list): the section of the mcnp output for this tally

        Returns:
            None, but creates self.statistical_checks, a dictionary with
                the results of the statistical checks for this tally.
        """
        start_PATTERN = re.compile(r"\s+results of 10 statistical checks.+")
        value = None
        for n, line in enumerate(data):
            if start_PATTERN.match(line):
                value = data[n+6].split()
                pass_fail = data[n+7].split()
                break
        if value:       # if statistical checks have been found
            statistical_checks = {
                'mean_behaviour'    : {'value': value[1], 'pass': pass_fail[1]},
                'rel_err_value'     : {'value': value[2], 'pass': pass_fail[2]},
                'rel_err_decrease'  : {'value': value[3], 'pass': pass_fail[3]},
                'rel_err_dec_rate'  : {'value': value[4], 'pass': pass_fail[4]},
                'VoV_value'         : {'value': value[5], 'pass': pass_fail[5]},
                'VoV_decrease'      : {'value': value[6], 'pass': pass_fail[6]},
                'VoV_dec_rate'      : {'value': value[7], 'pass': pass_fail[7]},
                'FoM_value'         : {'value': value[8], 'pass': pass_fail[8]},
                'FoM_behaviour'     : {'value': value[9], 'pass': pass_fail[9]},
                'pdf_slope'         : {'value': value[10], 'pass': pass_fail[10]},
                }
        else:
            # print(f"Tally {self.tally_number} does not have statistical checks")
            statistical_checks = {
                'mean_behaviour'    : {'value': r"none", 'pass': "no"},
                'rel_err_value'     : {'value': r"none", 'pass': "no"},
                'rel_err_decrease'  : {'value': r"none", 'pass': "no"},
                'rel_err_dec_rate'  : {'value': r"none", 'pass': "no"},
                'VoV_value'         : {'value': r"none", 'pass': "no"},
                'VoV_decrease'      : {'value': r"none", 'pass': "no"},
                'VoV_dec_rate'      : {'value': r"none", 'pass': "no"},
                'FoM_value'         : {'value': r"none", 'pass': "no"},
                'FoM_behaviour'     : {'value': r"none", 'pass': "no"},
                'pdf_slope'         : {'value': r"none", 'pass': "no"},
            }
        return statistical_checks

    def get_passes(self):
        """Determine how many statistical checks were passed

        Returns:
            Passes (int): the number of checks passed (out of ten)
        """
        passes = 0
        if self.statistical_checks:
            for check in self.statistical_checks:
                if self.statistical_checks[check]['pass'] == 'yes':
                    passes += 1
        return passes

    def normalise_data(self):
        """Each subclass of Tally should have its own normalise_data method"""
        print("This Tally class appears not to have its own 'normalise_data() method.")
        raise NotImplementedError(f"The {self.__class__} subclass should have its own normalise_data method")

    def describe_object(self):
        """ Print a description of the Tally object to the terminal."""
        print()
        print(f"Tally {self.tally_number}")
        print(f"Tally type: {self.f_type}, {self.tally_type}")
        print(f"Particles: {self.particles}")
        print(f"Number of particles: {self.nps}")
        print(f"This tally is modified by dose functions {self.dose_functions[0]} and {self.dose_functions[1]}")
        print(f"Tally Results: ")
        for region in self.results:
            print(f"    {region['region']}:     Tally Result = {region['result']}      Variance = {region['variance']}")
        print()
        if 'statistical_checks' in dir(self):
            print("Statistical Check      Value      Pass?")
            for check in self.statistical_checks:
                print(f"{check + ':' : <21}  {self.statistical_checks[check]['value'] : <11} {self.statistical_checks[check]['pass'] : <3}")
            print(f"{self.passes} out of 10 statistical checks passed.")
            print()
        else:
            print("This tally does not have statistical checks.")


class F2Tally(Tally):
    def __init__(self, data):
        super().__init__(data)
        gv.F2_tallies[self.particles].append(self)

    def get_results(self, data):
        """
            Gets the tally results from the mcnp output file
            Args: self: the object, data: the mcnp output tally section
            Returns: regions: a dictionary holding the results for that tally
        """
        results = {}
        for num, line in enumerate(data):
            if "surface:" in line:
                results["surface"] = int(data[num].split()[1])
                results["areas"] = float(data[num+1].strip())
            elif "surface  " in line:
                results["result"] = float(data[num+1].split()[0])
                results["variance"] = float(data[num+1].split()[1])
                break
        return results

    def normalise_data(self):
        """
            Applies normalisation factor, which can be given as an argument to the core script.
            Args: self, scaling_factor: a float or int by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        self.results["result"] *= gv.scaling_factor


class F4Tally(Tally):
    def __init__(self, data):
        super().__init__(data)
        gv.F4_tallies[self.particles].append(self)

    def get_results(self, data):
        """Get the tally results from the mcnp output file

        Args:
            data (list): the section of the MCNP output file for this tally

        Returns:
            regions (dict): The results for this tally
        """
        results = []
        for num, line in enumerate(data):
            if (" cell  " in line) or (" surface  " in line):
                results.append({
                    "region": line.strip().capitalize(),
                    "result": float(data[num+1].split()[0]),
                    "variance": float(data[num+1].split()[1])
                })
        return results

    def normalise_data(self):
        """
            Applies scaling factor, which can be given as an argument to the core script.
            Args: self, scaling_factor: a float or int by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        for num, region in enumerate(self.results):
            region['result'] *= gv.scaling_factor


class F5Tally(Tally):
    def __init__(self, data):
        super().__init__(data)
        gv.F5_tallies[self.particles].append(self)

    def get_results(self, data):
        """
            Gets the tally results from the mcnp output file
            Args: self: the object, data: the mcnp output tally section
            Returns: regions: a dictionary holding the results for that tally
        """
        results = {}
        for num, line in enumerate(data):
            if "detector located at" in line:
                results["x"] = float(line[28:40])
                results["y"] = float(line[40:52])
                results["z"] = float(line[52:64])
                results["result"] = float(data[num+1].split()[0])
                results["variance"] = float(data[num+1].split()[1])
                break
        return results

    def describe_object(self):
        """ Prints a description of the Tally object to the terminal."""
        print()
        print(f"Tally {self.tally_number}")
        print(f"Tally type: {self.f_type}, {self.tally_type}")
        print(f"Particles: {self.particles}")
        print(f"Number of particles: {self.nps}")
        print(f"This tally is modified by dose functions {self.dose_functions[0]} and {self.dose_functions[1]}")
        print(f"Tally Results: ")
        print(f"   Position: {self.results['x']},{self.results['y']},{self.results['z']}   Tally Result = {self.results['result']}      Variance = {self.results['variance']}")
        print()
        if 'statistical_checks' in dir(self):
            print("Statistical Check      Value      Pass?")
            for check in self.statistical_checks:
                print(f"{check + ':' : <21}  {self.statistical_checks[check]['value'] : <11} {self.statistical_checks[check]['pass'] : <3}")
            print(f"{self.passes} out of 10 statistical checks passed.")
            print()
        else:
            print("This tally does not have statistical checks.")

    def normalise_data(self):
        """
            Applies normalisation factor, which can be given as an argument to the core script.
            Args: self, scaling_factor: a float or int by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        self.results["result"] *= gv.scaling_factor


############################################################
#  End of Tally class                                      #
############################################################


def get_tallies(data):
    """Find the tally sections in the MCNP output.

    Args:
        data (list): the mcnp output

    Returns:
        None, but creates Tally class objects.
    """
    PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
    PATTERN_tally_start = re.compile(r'^\s*1tally\s+\d+\s+nps.+')
    PATTERN_tally_end = re.compile(r'(^\s*(1tally|1status).+)')
    for n, line in enumerate(data):
        if PATTERN_run_terminated.match(line):
            terminated = n
            for m, new_line in enumerate(data[terminated:], start=terminated):
                if PATTERN_tally_start.match(new_line):
                    tally_start = m
                    for o, other_line in enumerate(data[tally_start+1:], start=tally_start+1):
                        if PATTERN_tally_end.match(other_line):
                            tally_end = o
                            break
                    tally_data = data[tally_start:tally_end]
                    tally_type = tally_data[1].split()[2]
                    # Create tally object
                    if tally_type == "2":
                        F2Tally(tally_data)
                    if tally_type == "4":
                        F4Tally(tally_data)
                    if tally_type == "5":
                        F5Tally(tally_data)
            break



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""
    This module is the base of the SCALE to HTML converter section of Eddy.
"""

# local imports
from . import scale_tallies, mixtures, scale_global_variables as gv, scale_html_writer


def read_file(filename):
    """Read in the SCALE output file.

    Args:
        filename (str): The file path of the SCALE .out file

    Returns:
        data (list): The lines from the output file
    """
    with open(filename, 'r') as file:
        data = file.readlines()
    return data


def get_runtime(data):
    """Find the date and time that the SCALE case was run

    Args:
        data(list): The SCALE output data

    Returns:
        A tuple containing
            rundate (str): the date the case was run
            runtime (str): the time the case was run
    """
    for line in data:
        if "Job started on" in line:
            rundate = line.split()[6]
            # rearrange date from m/d/y to y/m/d
            y = rundate.split('/')[2]
            m = rundate.split('/')[1]
            d = rundate.split('/')[0]
            rundate = f"{y}/{m}/{d}"
            # time is fine as h/m/s
            runtime = line.split()[7]
            return rundate, runtime


def get_input(data):
    """Get the SCALE input deck from the output file

    Args:
        data (list): The SCALE output data

    Returns:
        input_data (list): The lines from the output containing the input data
    """
    for n, line in enumerate(data):
        if "Input Data:" in line:
            for m, other_line in enumerate(data[n+1:], start=n+1):
                if other_line.lower() == 'end\n':
                    input_data = data[n+2:m+1]
                    return input_data


def parse_output(data):
    """Crete various objects from the output data by calling other modules

    Args:
        data (list): The SCALE output data

    Returns:
        None
    """
    tally_data = scale_tallies.get_tally_data(data)
    gv.rundate, gv.runtime = get_runtime(data)
    scale_tallies.create_tallies(tally_data)
    scale_tallies.Tally.scale_results()
    mixture_data = mixtures.get_mixture_data(data)
    mixtures.create_mixtures(mixture_data)
    gv.scale_input = get_input(data)


def main(filename, scaling_factor=1):
    """Provide an entry to the SCALE converter and call the other functions in this module

    Args:
        filename (str): The SCALE output file path
        scaling_factor (float): The scaling factor to multiply the results by

    Returns:
        None
    """
    gv.filename = filename
    gv.scaling_factor = scaling_factor
    output_data = read_file(filename)
    parse_output(output_data)
    scale_html_writer.main(filename)




""" This module holds global (project-level) variables for the MCNP output to HTML converter. """

f = None
scaling_factor = 1
date_time = {}
rundate = None
runtime = None
parameters = {}
cell_list = []
tally_list = []
f_types = []
F2_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
F4_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
F5_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
tally_types = {}
warnings = []
comments = []
duplicate_surfaces = []
particle_list = []
k_effective = {}
ctme = None
nps = None
mcnp_input = None
crit_case = False
cycles = {}



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""This module contains all the code relating to SCALE mixtures."""
# import from standard library:
import re
# local imports:
from . import scale_global_variables as gv


class Mixture:
    """Each mixture in the SCALE case is represented by a Mixture object"""
    def __init__(self, data):
        self.number = data[0].split()[2]
        self.density = data[0].split()[5]
        try:
            self.temperature = data[0].split()[8] + 'K'
        except IndexError:
            self.temperature = "Not given in output file"
        self.isotopes = {}
        for line in data[2:]:
            nuclide = line.split()[0]
            self.isotopes[nuclide] = {}
            self.isotopes[nuclide]['nuclide'] = int(nuclide)
            self.isotopes[nuclide]['atom-density'] = float(line.split()[1])
            self.isotopes[nuclide]['weight fraction'] = float(line.split()[2])
            self.isotopes[nuclide]['z-number'] = int(line.split()[3])
            self.isotopes[nuclide]['atomic weight'] = float(line.split()[4])
            self.isotopes[nuclide]['title'] = line.split()[5].capitalize()
            self.isotopes[nuclide]['temperature'] = float(line.split()[6])
        gv.mixture_list.append(self)


def get_mixture_data(output_data):
    """Get the part of the SCALE output concerning mixtures

    Args:
        output_data (list): The SCALE output data

    Returns:
        The lines from output_data concerning mixtures
    """
    pattern_mix = re.compile(r'^\s*mixing table\s+')
    for n, line in enumerate(output_data):
        if pattern_mix.match(line):
            for m, other_line in enumerate(output_data[n+1:], start=n+1):
                if "Cross section" in other_line or "***********" in other_line:
                    return output_data[n:m]


def create_mixtures(mix_data):
    """Create Mixture objects from the SCALE output data

    Args:
        mix_data (list): The lines from output_data concerning mixtures

    Returns:
        None, but creates Mixture objects
    """
    for n, line in enumerate(mix_data):
        if "mixture = " in line:
            for m, other_line in enumerate(mix_data[n+1:], start=n+1):
                if other_line == "\n":
                    mix = mix_data[n:m]
                    Mixture(mix)
                    break

